
import type { Func, Params } from '@js/types';

export interface MemoOptions<T extends Func> {
	/**
	 * The function to call and cache results from
	 */
	func: T;

	/**
	 * Time-to-live for records in the cache. After this amount of time,
	 * a value is considered "stale", and by default, will be evicted
	 * from the cache.
	 * 
	 * This value can be a number (ms), or a function which receives the
	 * params and result of the call and returns a number for each record.
	 */
	ttl: number | MemoTTLFunc<T>;

	/**
	 * Custom function for generating a cache key. If not provided, a
	 * built-in default will be used instead. (At the moment, that
	 * default is `JSON.stringify()`, which is not guranteed to be
	 * stable or particularly fast; Providing a better implementation
	 * for most functions is recommended)
	 */
	key?: MemoKeyFunc<T>;

	/**
	 * Callback to run when evicting a record from the cache (usually
	 * because it expired)
	 */
	on_evict?: MemoOnEvict<T>;
}

export interface MemoKeyFunc<T extends Func> {
	(params: Params<T>) : string;
}

export interface MemoTTLFunc<T extends Func> {
	(params: Params<T>, result: ReturnType<T>) : number;
}

export interface MemoOnEvict<T extends Func> {
	(evicted: MemoStoredResult<T>) : any;
}

export interface MemoStoredResult<T extends Func> {
	key: string;
	created: number;
	expires: number;
	result: ReturnType<T>;
	hits: number;
	timer: ReturnType<typeof setTimeout>;
}

export type MemoizedFunc<T extends Func> = T & {
	invalidate(...params: Params<T>) : void;
};

export function memo<T extends Func>(opts: MemoOptions<T>) : MemoizedFunc<T> {
	const get_key = opts.key ?? default_key;
	const storage: Record<string, MemoStoredResult<T>> = Object.create(null);

	function memoized(...args: Params<T>) : ReturnType<T> {
		const key = get_key(args);

		if (storage[key]) {
			storage[key].hits++;
			return storage[key].result;
		}

		const result = opts.func(...args);
		store_record(key, args, result);
		return result;
	}

	function evict(key: string) {
		if (storage[key]) {
			const evicted = storage[key];
			delete storage[key];

			if (evicted.timer) {
				clearTimeout(evicted.timer);
				evicted.timer = null;
			}

			if (opts.on_evict) {
				opts.on_evict(evicted);
			}
		}
	}

	function store_record(key: string, params: Params<T>, result: ReturnType<T>) {
		const record: MemoStoredResult<T> = {
			key,
			created: Date.now(),
			expires: null,
			result,
			hits: 0,
			timer: null,
		};

		const ttl = typeof opts.ttl === 'function'
			? opts.ttl(params, result)
			: opts.ttl;

		record.expires = record.created + ttl;
		record.timer = setTimeout(() => evict(record.key), ttl);
	}

	memoized.invalidate = function invalidate(...params: Params<T>) {
		const key = get_key(params);
		evict(key);
	};
	
	return memoized as unknown as MemoizedFunc<T>;
}

function default_key(...params: any[]) : string {
	return JSON.stringify(params);
}
